"""
Connectors wrap the details of communicating with different Bitcoin clients and implementations.
"""
import io
import logging
import time
from abc import abstractmethod

import bitcoin.rpc
import requests
from bitcoin.core import CTransaction
from cert_core import Chain
from pycoin.coins.bitcoin.Tx import Tx
from pycoin.encoding.hexbytes import b2h, b2h_rev, h2b, h2b_rev
from pycoin.services import providers
from pycoin.services.chain_so import ChainSoProvider
from pycoin.services.insight import InsightProvider
from pycoin.services.providers import service_provider_methods
from pycoin.coins.bitcoin.Spendable import Spendable

import cert_issuer.config
from cert_issuer import helpers
from cert_issuer.errors import BroadcastError

try:
    from urllib2 import urlopen, HTTPError
    from urllib import urlencode
except ImportError:
    from urllib.request import urlopen, HTTPError
    from urllib.parse import urlencode

BROADCAST_RETRY_INTERVAL = 30
MAX_BROADCAST_ATTEMPTS = 3


def to_hex(transaction):
    s = io.BytesIO()
    transaction.stream(s)
    tx_as_hex = b2h(s.getvalue())
    return tx_as_hex

class BlockcypherProvider(object):
    """
    Note that this needs an API token
    """

    def __init__(self, base_url, api_token=None):
        self.base_url = base_url
        self.api_token = api_token
        
        if self.api_token:
            logging.info('Blockcypher at ' + base_url + ' is configured with an API token')

    def broadcast_tx(self, tx):
        hextx = to_hex(tx)
        broadcast_url = self.base_url + '/txs/push'
        if self.api_token:
            broadcast_url += '?token=' + self.api_token
        response = requests.post(broadcast_url, json={'tx': hextx})
        if int(response.status_code) == 201:
            tx_id = response.json().get('tx', None)
            tx_hash = tx_id.get('hash', None)
            return tx_hash
        logging.error('Error broadcasting the transaction through the Blockcypher API. Error msg: %s', response.text)
        raise BroadcastError(response.text)

    def spendables_for_address(self, address):
        """
        Return a list of Spendable objects for the
        given bitcoin address.
        """
        logging.info('trying to get spendables from blockcypher')
        spendables = []
        url_append = '?unspentOnly=true&includeScript=true'
        if self.api_token:
            url_append += '&token=' + self.api_token
        url = self.base_url + '/addrs/' + address + url_append
        response = requests.get(url)
        if int(response.status_code) == 200:
            for txn in response.json().get('txrefs', []):
                coin_value = txn.get('value')
                script = h2b(txn.get('script'))
                previous_hash = h2b_rev(txn.get('tx_hash'))
                previous_index = txn.get('tx_output_n')
                spendables.append(Spendable(coin_value, script, previous_hash, previous_index))
        return spendables

class BlockstreamProvider(object):
    def __init__(self, base_url_free, base_url_enterprise, client_id=None, client_secret=None):
        if base_url_enterprise and client_id and client_secret:
            logging.info('Blockstream provider for ' + base_url_enterprise + ' configured for enterprise tier')
            self.base_url = base_url_enterprise
        elif base_url_enterprise:
            logging.warning('Blockstream enterprise base URL is set, but CLIENT_ID and/or CLIENT_SECRET are missing. Defaulting to free tier.')
            self.base_url = base_url_free
            logging.info('Blockstream provider for ' + base_url_free + ' configured for free tier')
        else:
            self.base_url = base_url_free
            logging.info('Blockstream provider for ' + base_url_free + ' configured for free tier')
        
        self.client_id = client_id
        self.client_secret = client_secret

    def get_access_token(self):
        if not self.client_id or not self.client_secret:
            logging.warning('Client ID or Client Secret is not set.')
            return None
        token_url = 'https://login.blockstream.com/realms/blockstream-public/protocol/openid-connect/token'
        
        payload = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'client_credentials',
            'scope': 'openid'
        }
        
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        
        try:
            response = requests.post(token_url, data=payload, headers=headers)
            response.raise_for_status()  # Raise an error for bad responses
            
            # Assuming the token is returned in JSON format under key 'access_token'
            access_token = response.json().get('access_token')
            return access_token
        except requests.exceptions.RequestException as e:
            logging.error('Error obtaining access token: %s', str(e))
            return None

    def broadcast_tx(self, tx):
        hextx = to_hex(tx)
        broadcast_url = self.base_url + '/tx'
        
        access_token = self.get_access_token()

        if access_token:
            logging.info('Using enterprise-tier Blockstream to broadcast transaction')
            headers = {
                'Authorization': f'Bearer {access_token}',
            }
        else:
            logging.info('Using free-tier Blockstream to broadcast transaction.')
            headers = {}
        
        response = requests.post(broadcast_url, data=hextx, headers=headers)
        
        if int(response.status_code) == 200:
            tx_id = response.text
            return tx_id
        logging.error('Error broadcasting the transaction through the Blockstream API. Error msg: %s', response.text)
        raise BroadcastError(response.text)
    
    def spendables_for_address(self, address):
        logging.info('Trying to get spendables from Blockstream')
        spendables = []

        # Default headers
        headers = {}

        # Determine if we should use authenticated access
        access_token = None
        if self.client_id and self.client_secret:
            access_token = self.get_access_token()
            if access_token:
                logging.info('Using enterprise-tier Blockstream to get spendables.')
                headers['Authorization'] = f'Bearer {access_token}'
            else:
                logging.warning('Failed to get access token; falling back to free-tier Blockstream for spendables.')
        else:
            logging.info('Using free-tier Blockstream to get spendables.')

        # Step 1: Get UTXOs for the address
        url = f"{self.base_url}/address/{address}/utxo"
        response = requests.get(url, headers=headers)

        if int(response.status_code) == 200:
            utxos = response.json()
            for utxo in utxos:
                coin_value = utxo.get('value')
                previous_hash = h2b_rev(utxo.get('txid'))
                previous_index = utxo.get('vout')

                # Step 2: Get full transaction to retrieve scriptPubKey
                tx_url = f"{self.base_url}/tx/{utxo['txid']}/hex"
                tx_response = requests.get(tx_url, headers=headers)

                if int(tx_response.status_code) == 200:
                    tx_hex = tx_response.text
                    tx = Tx.from_hex(tx_hex)
                    script = tx.txs_out[utxo['vout']].script

                    spendables.append(Spendable(coin_value, script, previous_hash, previous_index))
                else:
                    logging.error(f"Error fetching full transaction {utxo['txid']} from Blockstream.")
        else:
            logging.error(f"Error fetching UTXOs for address {address}: {response.text}")

        return spendables


class BitcoindConnector(object):
    def __init__(self, netcode):
        self.netcode = netcode

    def broadcast_tx(self, transaction):
        as_hex = transaction.as_hex()
        transaction = CTransaction.deserialize(h2b(as_hex))
        tx_id = bitcoin.rpc.Proxy().sendrawtransaction(transaction)
        # reverse endianness for bitcoind
        return b2h_rev(tx_id)

    def spendables_for_address(self, address):
        """
        Converts to pycoin Spendable type
        :param address:
        :return: list of Spendables
        """
        unspent_outputs = bitcoin.rpc.Proxy().listunspent(addrs=[address])
        logging.debug('spendables_for_address %s', address)

        spendables = []
        for unspent in unspent_outputs:
            coin_value = unspent.get('amount', 0)
            outpoint = unspent.get('outpoint')
            script = unspent.get('scriptPubKey')
            previous_hash = outpoint.hash
            previous_index = outpoint.n
            spendables.append(Spendable(coin_value, script, previous_hash, previous_index))
        return spendables


class ServiceProviderConnector(object):
    @abstractmethod
    def get_balance(self, address):
        pass

    def broadcast_tx(self, tx):
        pass


class MockServiceProviderConnector(ServiceProviderConnector):
    def get_balance(self, address):
        pass

    def broadcast_tx(self, tx):
        pass


class BitcoinServiceProviderConnector(ServiceProviderConnector):
    def __init__(self, bitcoin_chain, bitcoind=False):
        self.bitcoin_chain = bitcoin_chain
        self.bitcoind = bitcoind

    def spendables_for_address(self, bitcoin_address):
        for m in service_provider_methods('spendables_for_address',
                                          get_providers_for_chain(self.bitcoin_chain, self.bitcoind)):
            try:
                logging.debug('m=%s', m)
                spendables = m(bitcoin_address)
                return spendables
            except Exception as e:
                logging.warning(e)
                pass
        return []

    def get_unspent_outputs(self, address):
        """
        Get unspent outputs at the address
        :param address:
        :return:
        """
        logging.debug('get_unspent_outputs for address=%s', address)
        spendables = self.spendables_for_address(bitcoin_address=address)
        if spendables:
            return sorted(spendables, key=lambda x: hash(x.coin_value))
        return None

    def get_balance(self, address):
        """
        Get balance available to spend at the address
        :param address:
        :return:
        """
        spendables = self.get_unspent_outputs(address)
        if not spendables:
            logging.warning('address %s has a balance of 0', address)
            return 0

        balance = sum(s.coin_value for s in spendables)
        return balance

    def broadcast_tx(self, tx):
        """
        Broadcast the transaction through the configured set of providers

        :param tx:
        :return:
        """
        return BitcoinServiceProviderConnector.broadcast_tx_with_chain(tx, self.bitcoin_chain, self.bitcoind)

    @staticmethod
    def broadcast_tx_with_chain(tx, bitcoin_chain, bitcoind=False):
        """
        Broadcast the transaction through the configured set of providers

        :param tx:
        :param bitcoin_chain:
        :return:
        """
        last_exception = None
        final_tx_id = None

        # Unlike other providers, we want to broadcast to all available apis
        for attempt_number in range(0, MAX_BROADCAST_ATTEMPTS):
            for method_provider in service_provider_methods('broadcast_tx',
                                                            get_providers_for_chain(bitcoin_chain, bitcoind)):
                try:
                    tx_id = method_provider(tx)
                    if tx_id:
                        # We have to check to make sure that the tx_id at least kind of looks like a sha256 hash string.
                        if len(tx_id) != 64:
                            if final_tx_id:
                                # If we already found a final_tx_id, then some other provider got it, but we should
                                # log the bad provider.
                                logging.error("Ignoring invalid tx_id %s from provider %s since valid" +
                                    "tx_id %s already exists (another provider already sent the transaction).  " +
                                    "The current provider might be broken!  Bad txid!",
                                    tx_id, str(method_provider), final_tx_id)
                                continue
                            else:
                                logging.error("Provider %s returned an invalid tx_id: %s.  " +
                                    "Aborting further retries.  Bad txid!",
                                    str(method_provider), tx_id)
                                raise BroadcastError("Invalid tx_id received: " + tx_id)
                        
                        logging.info('Broadcasting succeeded with method_provider=%s, txid=%s', str(method_provider),
                                     tx_id)
                        if final_tx_id and final_tx_id != tx_id:
                            logging.error(
                                'This should never happen; fail and investigate if it does. Got conflicting tx_ids=%s and %s. Hextx=%s',
                                final_tx_id, tx_id, tx.as_hex())
                            raise Exception('Got conflicting tx_ids.')
                        final_tx_id = tx_id
                except Exception as e:
                    logging.warning('Caught exception trying provider %s. Trying another. Exception=%s',
                                    str(method_provider), e)
                    last_exception = e
            # At least 1 provider succeeded, so return
            if final_tx_id:
                return final_tx_id
            else:
                logging.warning('Broadcasting failed. Waiting before retrying. This is attempt number %d',
                                attempt_number)
                time.sleep(BROADCAST_RETRY_INTERVAL)
        logging.error('Failed broadcasting through all providers')
        logging.error(last_exception, exc_info=True)
        raise BroadcastError(last_exception)


# configure api tokens
config = cert_issuer.config.CONFIG
blockcypher_token = None if config is None else config.blockcypher_api_token
blockstream_client_id = None if config is None else config.blockstream_client_id
blockstream_client_secret = None if config is None else config.blockstream_client_secret

PYCOIN_BTC_PROVIDERS = "chain.so"  # blockchain.info blockcypher.com
PYCOIN_XTN_PROVIDERS = ""  # chain.so

# initialize connectors
connectors = {}

# configure mainnet providers
provider_list = providers.providers_for_config_string(
    PYCOIN_BTC_PROVIDERS,
    helpers.to_pycoin_chain(Chain.bitcoin_mainnet)
)
provider_list.append(
    BlockstreamProvider(
        'https://blockstream.info/api',
        'https://enterprise.blockstream.info/api',
        blockstream_client_id,
        blockstream_client_secret
    )
)
provider_list.append(
    BlockcypherProvider(
        'https://api.blockcypher.com/v1/btc/main',
        blockcypher_token
    )
)
provider_list.append(
    InsightProvider(
        netcode=helpers.to_pycoin_chain(Chain.bitcoin_mainnet)
    )
)
#provider_list.append(
#    ChainSoProvider(
#        netcode=helpers.to_pycoin_chain(Chain.bitcoin_mainnet)
#    )
#)

connectors[Chain.bitcoin_mainnet] = provider_list

# configure testnet providers
xtn_provider_list = providers.providers_for_config_string(
    PYCOIN_XTN_PROVIDERS,
    helpers.to_pycoin_chain(Chain.bitcoin_testnet)
)
#xtn_provider_list.append(
#    ChainSoProvider(
#        netcode=helpers.to_pycoin_chain(Chain.bitcoin_testnet)
#    )
#)

xtn_provider_list.append(
    BlockstreamProvider(
        'https://blockstream.info/testnet/api',
        'https://enterprise.blockstream.info/testnet/api',
        blockstream_client_id,
        blockstream_client_secret
    )
)

xtn_provider_list.append(
    BlockcypherProvider(
        'https://api.blockcypher.com/v1/btc/test3',
        blockcypher_token
    )
)

connectors[Chain.bitcoin_testnet] = xtn_provider_list

def get_providers_for_chain(chain, bitcoind=False):
    if bitcoind:
        return [BitcoindConnector(helpers.to_pycoin_chain(chain))]
    else:
        return connectors[chain]



import os
import shutil
import tempfile
import unittest
import json
from pathlib import Path

import importlib
from cert_issuer import helpers
from cert_issuer.errors import NoCertificatesFoundError

class TestPrepareIssuanceBatch(unittest.TestCase):
    def setUp(self):
        importlib.reload(helpers)
        
        self.temp_dir = tempfile.TemporaryDirectory()
        self.base_dir = Path(self.temp_dir.name)

        self.unsigned_dir = self.base_dir / "unsigned"
        self.signed_dir = self.base_dir / "signed"
        self.blockchain_dir = self.base_dir / "blockchain"
        self.work_dir = self.base_dir / "work"

        self.unsigned_dir.mkdir()
        (self.unsigned_dir / "cert1.json").write_text(json.dumps({"id": 1}))
        (self.unsigned_dir / "cert2.json").write_text(json.dumps({"id": 2}))
        (self.unsigned_dir / "not_a_cert.txt").write_text("ignore me")

    def tearDown(self):
        self.temp_dir.cleanup()

    def test_prepares_and_identifies_certificates(self):
        metadata = helpers.prepare_issuance_batch(
            unsigned_certs_dir=str(self.unsigned_dir),
            signed_certs_dir=str(self.signed_dir),
            blockchain_certs_dir=str(self.blockchain_dir),
            work_dir=str(self.work_dir)
        )

        # Assert we got the expected UIDs
        self.assertIn("cert1", metadata)
        self.assertIn("cert2", metadata)
        self.assertNotIn("not_a_cert", metadata)
        self.assertEqual(len(metadata), 2)

        # Ensure metadata has correct file paths
        for uid, meta in metadata.items():
            self.assertTrue(str(meta.unsigned_cert_file_name).endswith(f"{uid}.json"))
            self.assertTrue(str(meta.signed_cert_file_name).endswith(f"{uid}.json"))
            self.assertTrue(str(meta.blockchain_cert_file_name).endswith(f"{uid}.json"))
            self.assertTrue(str(meta.final_blockchain_cert_file_name).endswith(f"{uid}.json"))

    def test_raises_if_no_certificates(self):
        # Clean out the unsigned cert dir and re-test
        for file in self.unsigned_dir.glob("*"):
            file.unlink()

        with self.assertRaises(NoCertificatesFoundError):
            helpers.prepare_issuance_batch(
                unsigned_certs_dir=str(self.unsigned_dir),
                signed_certs_dir=str(self.signed_dir),
                blockchain_certs_dir=str(self.blockchain_dir),
                work_dir=str(self.work_dir)
            )


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import os.path
import sys

PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

if (__package__ is None or __package__ == '') and not hasattr(sys, 'frozen'):
    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))


def cert_issuer_main(args=None):
    from cert_issuer import config
    parsed_config = config.get_config()
    from cert_issuer import issue_certificates
    issue_certificates.main(parsed_config)


if __name__ == '__main__':
    cert_issuer_main()



import logging
import os
import time

import requests

from cert_issuer.models import SecretManager


class FileSecretManager(SecretManager):
    def __init__(self, signer, path_to_secret, safe_mode=True, issuing_address=None):
        super().__init__(signer)
        self.path_to_secret = path_to_secret
        self.safe_mode = safe_mode
        self.issuing_address = issuing_address

    def start(self):
        if self.safe_mode:
            check_internet_off(self.path_to_secret)
        else:
            logging.warning(
                'app is configured to skip the wifi check when the USB is plugged in. Read the documentation to'
                ' ensure this is what you want, since this is less secure')

        self.wif = import_key(self.path_to_secret)

    def stop(self):
        self.wif = None
        if self.safe_mode:
            check_internet_on(self.path_to_secret)
        else:
            logging.warning(
                'app is configured to skip the wifi check when the USB is plugged in. Read the documentation to'
                ' ensure this is what you want, since this is less secure')


class FinalizableSigner(object):
    def __init__(self, secret_manager):
        self.secret_manager = secret_manager

    def __enter__(self):
        logging.info('Starting finalizable signer')
        self.secret_manager.start()
        return self.secret_manager

    def __exit__(self, exc_type, exc_val, exc_tb):
        logging.info('Stopping finalizable signer')
        self.secret_manager.stop()


def import_key(secrets_file_path):
    with open(secrets_file_path) as key_file:
        key = key_file.read().strip()
    return key


def internet_on():
    """Pings Google to see if the internet is on. If online, returns true. If offline, returns false."""
    try:
        requests.get('http://google.com')
        return True
    except requests.exceptions.RequestException:
        return False


def check_internet_off(secrets_file_path):
    """If internet off and USB plugged in, returns true. Else, continues to wait..."""
    while True:
        if internet_on() is False and os.path.exists(secrets_file_path):
            break
        else:
            print("Turn off your internet and plug in your USB to continue...")
            time.sleep(10)
    return True


def check_internet_on(secrets_file_path):
    """If internet on and USB unplugged, returns true. Else, continues to wait..."""
    while True:
        if internet_on() is True and not os.path.exists(secrets_file_path):
            break
        else:
            print("Turn on your internet and unplug your USB to continue...")
            time.sleep(10)
    return True



import unittest
import mock

from cert_issuer.signer import FinalizableSigner


class TestSigner(unittest.TestCase):
    def test_finalizable_signer(self):
        mock_sm = mock.Mock()

        with FinalizableSigner(mock_sm) as fs:
            mock_sm.start.assert_called()

        mock_sm.stop.assert_called()


if __name__ == '__main__':
    unittest.main()



import logging
import random

from pycoin.encoding.hexbytes import b2h

from cert_issuer.blockchain_handlers.bitcoin import tx_utils
from cert_issuer.config import ESTIMATE_NUM_INPUTS, V2_NUM_OUTPUTS
from cert_issuer.errors import InsufficientFundsError
from cert_issuer.models import TransactionCreator, TransactionHandler
from cert_issuer.signer import FinalizableSigner


class TransactionV2Creator(TransactionCreator):
    def estimate_cost_for_certificate_batch(self, tx_cost_constants, num_inputs=ESTIMATE_NUM_INPUTS):
        total = tx_utils.calculate_tx_fee(tx_cost_constants, num_inputs, V2_NUM_OUTPUTS)
        return total

    def create_transaction(self, tx_cost_constants, issuing_address, inputs, op_return_value):
        fee = tx_utils.calculate_tx_fee(tx_cost_constants, len(inputs), V2_NUM_OUTPUTS)
        transaction = tx_utils.create_trx(
            op_return_value,
            fee,
            issuing_address,
            [],
            inputs)

        return transaction


class BitcoinTransactionHandler(TransactionHandler):
    def __init__(self, connector, tx_cost_constants, secret_manager, issuing_address, prepared_inputs=None,
                 transaction_creator=TransactionV2Creator()):
        self.connector = connector
        self.tx_cost_constants = tx_cost_constants
        self.secret_manager = secret_manager
        self.issuing_address = issuing_address
        self.prepared_inputs = prepared_inputs
        self.transaction_creator = transaction_creator

    def ensure_balance(self):
        # ensure the issuing address has sufficient balance
        balance = self.connector.get_balance(self.issuing_address)

        transaction_cost = self.transaction_creator.estimate_cost_for_certificate_batch(self.tx_cost_constants)
        logging.info('Total cost will be %d satoshis', transaction_cost)

        if transaction_cost > balance:
            error_message = 'Please add {} satoshis to the address {}'.format(
                transaction_cost - balance, self.issuing_address)
            logging.error(error_message)
            raise InsufficientFundsError(error_message)

    def issue_transaction(self, blockchain_bytes):
        op_return_value = b2h(blockchain_bytes)
        prepared_tx = self.create_transaction(blockchain_bytes)
        signed_tx = self.sign_transaction(prepared_tx)
        self.verify_transaction(signed_tx, op_return_value)
        txid = self.broadcast_transaction(signed_tx)
        # this logging is already done in issuer
        # logging.info('Broadcast transaction with txid %s', txid)
        return txid

    def create_transaction(self, op_return_bytes):
        if self.prepared_inputs:
            inputs = self.prepared_inputs
        else:
            spendables = self.connector.get_unspent_outputs(self.issuing_address)
            if not spendables:
                error_message = 'No money to spend at address {}'.format(self.issuing_address)
                logging.error(error_message)
                raise InsufficientFundsError(error_message)

            cost = self.transaction_creator.estimate_cost_for_certificate_batch(self.tx_cost_constants)
            current_total = 0
            inputs = []
            random.shuffle(spendables)
            for s in spendables:
                inputs.append(s)
                current_total += s.coin_value
                if current_total > cost:
                    break

        tx = self.transaction_creator.create_transaction(self.tx_cost_constants, self.issuing_address, inputs,
                                                         op_return_bytes)
        hex_tx = b2h(tx.serialize())
        logging.info('Unsigned hextx=%s', hex_tx)
        prepared_tx = tx_utils.prepare_tx_for_signing(hex_tx, inputs)
        return prepared_tx

    def sign_transaction(self, prepared_tx):
        with FinalizableSigner(self.secret_manager) as signer:
            signed_tx = signer.sign_transaction(prepared_tx)

        # log the actual byte count
        tx_byte_count = tx_utils.get_byte_count(signed_tx)
        logging.info('The actual transaction size is %d bytes', tx_byte_count)

        signed_hextx = signed_tx.as_hex()
        logging.info('Signed hextx=%s', signed_hextx)
        return signed_tx

    def verify_transaction(self, signed_tx, op_return_value):
        signed_hextx = signed_tx.as_hex()
        logging.info('Signed hextx=%s', signed_hextx)
        tx_utils.verify_transaction(signed_hextx, op_return_value)

    def broadcast_transaction(self, signed_tx):
        tx_id = self.connector.broadcast_tx(signed_tx)
        return tx_id



import logging
import os

from cert_core import BlockchainType
from cert_core import Chain, UnknownChainError

from cert_issuer.blockchain_handlers.bitcoin.connectors import BitcoinServiceProviderConnector, MockServiceProviderConnector
from cert_issuer.blockchain_handlers.bitcoin.signer import BitcoinSigner
from cert_issuer.blockchain_handlers.bitcoin.transaction_handlers import BitcoinTransactionHandler
from cert_issuer.certificate_handlers import CertificateBatchHandler, CertificateV2Handler, CertificateBatchWebHandler, CertificateWebV2Handler
from cert_issuer.merkle_tree_generator import MerkleTreeGenerator
from cert_issuer.models import MockTransactionHandler
from cert_issuer.signer import FileSecretManager

COIN = 100000000  # satoshis in 1 btc


class BitcoinTransactionCostConstants(object):
    def __init__(self, recommended_tx_fee=0.0006, min_per_output=0.0000275, satoshi_per_byte=250):
        self.recommended_tx_fee = recommended_tx_fee
        self.min_per_output = min_per_output
        self.satoshi_per_byte = satoshi_per_byte
        logging.info('Set cost constants to recommended_tx_fee=%f,min_per_output=%f,satoshi_per_byte=%d',
                     self.recommended_tx_fee, self.min_per_output, self.satoshi_per_byte)

    def get_minimum_output_coin(self):
        return self.min_per_output * COIN

    def get_recommended_fee_coin(self):
        return self.recommended_tx_fee * COIN


def initialize_signer(app_config):
    path_to_secret = os.path.join(app_config.usb_name, app_config.key_file)

    if app_config.chain.blockchain_type == BlockchainType.bitcoin:
        signer = BitcoinSigner(bitcoin_chain=app_config.chain)
    elif app_config.chain == Chain.mockchain:
        signer = None
    else:
        raise UnknownChainError(app_config.chain)
    secret_manager = FileSecretManager(signer=signer, path_to_secret=path_to_secret,
                                       safe_mode=app_config.safe_mode, issuing_address=app_config.issuing_address)
    return secret_manager

def instantiate_blockchain_handlers(app_config, file_mode=True):
    issuing_address = app_config.issuing_address
    chain = app_config.chain
    secret_manager = initialize_signer(app_config)

    if file_mode:
        certificate_batch_handler = CertificateBatchHandler(secret_manager=secret_manager,
                                                            certificate_handler=CertificateV2Handler(),
                                                            merkle_tree=MerkleTreeGenerator())
    else:
        certificate_batch_handler = CertificateBatchWebHandler(secret_manager=secret_manager,
                                                        certificate_handler=CertificateWebV2Handler(),
                                                        merkle_tree=MerkleTreeGenerator())
    if chain == Chain.mockchain:
        transaction_handler = MockTransactionHandler()
        connector = MockServiceProviderConnector()
    else:
        cost_constants = BitcoinTransactionCostConstants(app_config.tx_fee, app_config.dust_threshold,
                                                         app_config.satoshi_per_byte)
        connector = BitcoinServiceProviderConnector(chain, app_config.bitcoind)
        transaction_handler = BitcoinTransactionHandler(connector, cost_constants, secret_manager,
                                                        issuing_address=issuing_address)

    return certificate_batch_handler, transaction_handler, connector
